package cypher

import (
	"encoding/json"
	"errors"
	"strings"
	"testing"

	"git.science.uu.nl/GraphPolaris/query-conversion/entity"
	"github.com/stretchr/testify/assert"
)

func TestEmptyQueryConversion(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
		"return": {
			"entities": [],
			"relations": []
		},
		"entities": [],
		"relations": [],
		"limit": 5000
	}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	convertedResult, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)

	// Assert that the result and the expected result are the same
	correctConvertedResult := ``
	assert.Equal(t, correctConvertedResult, *convertedResult)
}

func TestEntityOneAttributeQuery(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
		"return": {
			"entities": [
				0
			],
			"relations": []
		},
		"entities": [
			{
				"type": "airports",
				"constraints": [
					{
						"attribute": "state",
						"value": "HI",
						"dataType": "string",
						"matchType": "exact"
					}
				]
			}
		],
		"relations": [],
		"limit": 5000
	}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	convertedResult, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)

	// Assert that the result and the expected result are the same
	correctConvertedResult := `MATCH (n0:airports)WHERE n0.state = "HI" RETURN n0`
	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
	assert.Equal(t, correctConvertedResult, cleanedResult)
}

func TestRelationWithConstraint(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
		"return": {
			"entities": [
				0
			],
			"relations": [
				0
			]
		},
		"entities": [
			{
				"type": "airports",
				"constraints": [
					{
						"attribute": "state",
						"value": "HI",
						"dataType": "string",
						"matchType": "exact"
					}
				]
			}
		],
		"relations": [
			{
				"type": "flights",
				"depth": {
					"min": 1,
					"max": 1
				},
				"entityFrom": 0,
				"entityTo": -1,
				"constraints": [
					{
						"attribute": "Day",
						"value": "15",
						"dataType": "int",
						"matchType": "EQ"
					}
				]
			}
		],
		"limit": 5000
	}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	convertedResult, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)

	// Assert that the result and the expected result are the same
	correctConvertedResult := `MATCH (n0:airports)WHERE n0.state = "HI" MATCH p0 = (n0)-[r0:flights*1..1]->()WHERE r0.Day = 15 RETURN n0,p0;`
	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
	assert.Equal(t, correctConvertedResult, cleanedResult)
}

// func TestModifierCountEntity(t *testing.T) {
// 	// Setup for test
// 	// Create query conversion service
// 	service := NewService()

// 	query := []byte(`{
// 		"return": {
// 			"entities": [
// 				0
// 			],
// 			"relations": []
// 		},
// 		"entities": [
// 			{
// 				"type": "airports",
// 				"constraints": [
// 					{
// 						"attribute": "state",
// 						"value": "HI",
// 						"dataType": "string",
// 						"matchType": "exact"
// 					}
// 				]
// 			}
// 		],
// 		"relations": [],
// 		"limit": 5000,
// 		"modifiers": [
// 			{
// 				"type": "COUNT",
// 				"selectedType": "entity",
// 				"id": 0,
// 				"attributeIndex": -1
// 			}
// 		]
// 	}`)

// 	// Unmarshall the incoming message into an IncomingJSONQuery object
// 	var JSONQuery entity.IncomingQueryJSON
// 	json.Unmarshal(query, &JSONQuery)

// 	convertedResult, err := service.ConvertQuery(&JSONQuery)

// 	// Assert that there is no error
// 	assert.NoError(t, err)

// 	// Assert that the result and the expected result are the same
// 	correctConvertedResult := `LET n0 = (FOR x IN airports FILTER x.state == "HI" RETURN x)RETURN LENGTH (n0)`
// 	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
// 	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
// 	assert.Equal(t, correctConvertedResult, cleanedResult)
// }

// func TestModifierCountEntityAttribute(t *testing.T) {
// 	// Setup for test
// 	// Create query conversion service
// 	service := NewService()

// 	query := []byte(`{
// 		"return": {
// 			"entities": [
// 				0
// 			],
// 			"relations": []
// 		},
// 		"entities": [
// 			{
// 				"type": "airports",
// 				"constraints": [
// 					{
// 						"attribute": "state",
// 						"value": "HI",
// 						"dataType": "string",
// 						"matchType": "exact"
// 					}
// 				]
// 			}
// 		],
// 		"relations": [],
// 		"limit": 5000,
// 		"modifiers": [
// 			{
// 				"type": "SUM",
// 				"selectedType": "entity",
// 				"id": 0,
// 				"attributeIndex": 0
// 			}
// 		]
// 	}`)

// 	// Unmarshall the incoming message into an IncomingJSONQuery object
// 	var JSONQuery entity.IncomingQueryJSON
// 	json.Unmarshal(query, &JSONQuery)

// 	convertedResult, err := service.ConvertQuery(&JSONQuery)

// 	// Assert that there is no error
// 	assert.NoError(t, err)

// 	// Assert that the result and the expected result are the same
// 	correctConvertedResult := `LET n0 = (FOR x IN airports FILTER x.state == "HI" RETURN x)RETURN SUM (n0[*].state)`
// 	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
// 	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
// 	assert.Equal(t, correctConvertedResult, cleanedResult)
// }

// func TestModifierCountRelation(t *testing.T) {
// 	// Setup for test
// 	// Create query conversion service
// 	service := NewService()

// 	query := []byte(`{
// 		"return": {
// 			"entities": [
// 				0
// 			],
// 			"relations": [
// 				0
// 			]
// 		},
// 		"entities": [
// 			{
// 				"type": "airports",
// 				"constraints": [
// 					{
// 						"attribute": "state",
// 						"value": "HI",
// 						"dataType": "string",
// 						"matchType": "exact"
// 					}
// 				]
// 			}
// 		],
// 		"relations": [
// 			{
// 				"type": "flights",
// 				"depth": {
// 					"min": 1,
// 					"max": 1
// 				},
// 				"entityFrom": 0,
// 				"entityTo": -1,
// 				"constraints": [
// 					{
// 						"attribute": "Day",
// 						"value": "15",
// 						"dataType": "int",
// 						"matchType": "EQ"
// 					}
// 				]
// 			}
// 		],
// 		"limit": 5000,
// 		"modifiers": [
// 			{
// 				"type": "COUNT",
// 				"selectedType": "relation",
// 				"id": 0,
// 				"attributeIndex": -1
// 			}
// 		]
// 	}`)

// 	// Unmarshall the incoming message into an IncomingJSONQuery object
// 	var JSONQuery entity.IncomingQueryJSON
// 	json.Unmarshal(query, &JSONQuery)

// 	convertedResult, err := service.ConvertQuery(&JSONQuery)

// 	// Assert that there is no error
// 	assert.NoError(t, err)

// 	// Assert that the result and the expected result are the same
// 	correctConvertedResult := `LET n0 = (FOR x IN airports FILTER x.state == "HI" RETURN x)LET r0 = (FOR x IN n0 FOR v, e, p IN 1..1 OUTBOUND x flights OPTIONS { uniqueEdges: "path" }FILTER p.edges[*].Day ALL == 15 RETURN DISTINCT p )RETURN LENGTH (unique(r0[*].edges[**]))`
// 	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
// 	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
// 	assert.Equal(t, correctConvertedResult, cleanedResult)
// }
// func TestModifierCountRelationAttribute(t *testing.T) {
// 	// Setup for test
// 	// Create query conversion service
// 	service := NewService()

// 	query := []byte(`{
// 		"return": {
// 			"entities": [
// 				0
// 			],
// 			"relations": [
// 				0
// 			]
// 		},
// 		"entities": [
// 			{
// 				"type": "airports",
// 				"constraints": [
// 					{
// 						"attribute": "state",
// 						"value": "HI",
// 						"dataType": "string",
// 						"matchType": "exact"
// 					}
// 				]
// 			}
// 		],
// 		"relations": [
// 			{
// 				"type": "flights",
// 				"depth": {
// 					"min": 1,
// 					"max": 1
// 				},
// 				"entityFrom": 0,
// 				"entityTo": -1,
// 				"constraints": [
// 					{
// 						"attribute": "Day",
// 						"value": "15",
// 						"dataType": "int",
// 						"matchType": "EQ"
// 					}
// 				]
// 			}
// 		],
// 		"limit": 5000,
// 		"modifiers": [
// 			{
// 				"type": "AVG",
// 				"selectedType": "relation",
// 				"id": 0,
// 				"attributeIndex": 0
// 			}
// 		]
// 	}`)

// 	// Unmarshall the incoming message into an IncomingJSONQuery object
// 	var JSONQuery entity.IncomingQueryJSON
// 	json.Unmarshal(query, &JSONQuery)

// 	convertedResult, err := service.ConvertQuery(&JSONQuery)

// 	// Assert that there is no error
// 	assert.NoError(t, err)

// 	// Assert that the result and the expected result are the same
// 	correctConvertedResult := `LET n0 = (FOR x IN airports FILTER x.state == "HI" RETURN x)LET r0 = (FOR x IN n0 FOR v, e, p IN 1..1 OUTBOUND x flights OPTIONS { uniqueEdges: "path" }FILTER p.edges[*].Day ALL == 15 RETURN DISTINCT p )RETURN AVG (r0[*].edges[**].Day)`
// 	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
// 	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
// 	assert.Equal(t, correctConvertedResult, cleanedResult)
// }

func TestRelationWithInOutConstraint(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
		"return": {
			"entities": [
				0,
				1
			],
			"relations": [
				0
			]
		},
		"entities": [
			{
				"type": "airports",
				"constraints": [
					{
						"attribute": "city",
						"value": "San Francisco",
						"dataType": "string",
						"matchType": "exact"
					}
				]
			},
			{
				"type": "airports",
				"constraints": [
					{
						"attribute": "state",
						"value": "HI",
						"dataType": "string",
						"matchType": "exact"
					}
				]
			}
		],
		"relations": [
			{
				"type": "flights",
				"depth": {
					"min": 1,
					"max": 3
				},
				"entityFrom": 1,
				"entityTo": 0,
				"constraints": [
					{
						"attribute": "Day",
						"value": "15",
						"dataType": "int",
						"matchType": "EQ"
					}
				]
			}
		],
		"limit": 5000
	}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	convertedResult, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)

	// Assert that the result and the expected result are the same
	correctConvertedResult := `MATCH (n1:airports)WHERE n1.state = "HI" MATCH p0 = (n1)-[r0:flights*1..3]->(n0)WHERE r0.Day = 15 RETURN n1,n0,p0;`
	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
	assert.Equal(t, correctConvertedResult, cleanedResult)
}

func TestTwoRelations(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
		"return": {
			"entities": [
				0,
				1,
				2
			],
			"relations": [
				0,
				1
			]
		},
		"entities": [
			{
				"type": "airports",
				"constraints": [
					{
						"attribute": "city",
						"value": "New York",
						"dataType": "string",
						"matchType": "exact"
					}
				]
			},
			{
				"type": "airports",
				"constraints": [
					{
						"attribute": "city",
						"value": "San Francisco",
						"dataType": "string",
						"matchType": "exact"
					}
				]
			},
			{
				"type": "airports",
				"constraints": [
					{
						"attribute": "state",
						"value": "HI",
						"dataType": "string",
						"matchType": "exact"
					}
				]
			}
		],
		"relations": [
			{
				"type": "flights",
				"depth": {
					"min": 1,
					"max": 3
				},
				"entityFrom": 2,
				"entityTo": 1,
				"constraints": [
					{
						"attribute": "Day",
						"value": "15",
						"dataType": "int",
						"matchType": "EQ"
					}
				]
			},
			{
				"type": "flights",
				"depth": {
					"min": 1,
					"max": 1
				},
				"entityFrom": 0,
				"entityTo": -1,
				"constraints": []
			}
		],
		"limit": 5000
	}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	convertedResult, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)

	// Assert that the result and the expected result are the same
	correctConvertedResult := `MATCH (n2:airports)WHERE n2.state = "HI" MATCH p0 = (n2)-[r0:flights*1..3]->(n1)WHERE r0.Day = 15 RETURN n2,n1,p0;MATCH (n0:airports)WHERE n0.city = "New York" MATCH p0 = (n0)-[r0:flights*1..1]->()RETURN n0,p0;`
	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
	assert.Equal(t, correctConvertedResult, cleanedResult)
}

func TestRelationWithOnlyToNode(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
			"return": {
				"entities": [
					0
				],
				"relations": [
					0
				]
			},
			"entities": [
				{
					"type": "airports",
					"constraints": [
						{
							"attribute": "city",
							"value": "San Francisco",
							"dataType": "string",
							"matchType": "exact"
						}
					]
				}
			],
			"relations": [
				{
					"type": "flights",
					"depth": {
						"min": 1,
						"max": 1
					},
					"entityFrom": -1,
					"entityTo": 0,
					"constraints": []
				}
			],
			"limit": 5000
		}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	convertedResult, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)

	// Assert that the result and the expected result are the same
	correctConvertedResult := `MATCH (n0:airports)WHERE n0.city = "San Francisco" MATCH p0 = (n0)-[r0:flights*1..1]->()RETURN n0,p0;`
	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
	assert.Equal(t, correctConvertedResult, cleanedResult)
}

func TestTooManyReturnEntities(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
			"return": {
				"entities": [
					0,
					1,
					2
				],
				"relations": [
					0
				]
			},
			"entities": [
				{
					"type": "airports",
					"constraints": [
						{
							"attribute": "city",
							"value": "San Francisco",
							"dataType": "string",
							"matchType": "exact"
						}
					]
				}
			],
			"relations": [
				{
					"type": "flights",
					"depth": {
						"min": 1,
						"max": 1
					},
					"entityFrom": -1,
					"entityTo": 0,
					"constraints": []
				}
			],
			"limit": 5000
		}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	_, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.Equal(t, errors.New("non-existing entity referenced in return"), err)
}

func TestTooManyReturnRelations(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
			"return": {
				"entities": [
					0
				],
				"relations": [
					0,
					1,
					2
				]
			},
			"entities": [
				{
					"type": "airports",
					"constraints": [
						{
							"attribute": "city",
							"value": "San Francisco",
							"dataType": "string",
							"matchType": "exact"
						}
					]
				}
			],
			"relations": [
				{
					"type": "flights",
					"depth": {
						"min": 1,
						"max": 1
					},
					"entityFrom": -1,
					"entityTo": 0,
					"constraints": []
				}
			],
			"limit": 5000
		}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	_, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.Equal(t, errors.New("non-existing relation referenced in return"), err)
}

func TestNegativeReturnEntities(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
			"return": {
				"entities": [
					0,
					-1
				],
				"relations": [
					0,
					1,
					2
				]
			},
			"entities": [
				{
					"type": "airports",
					"constraints": [
						{
							"attribute": "city",
							"value": "San Francisco",
							"dataType": "string",
							"matchType": "exact"
						}
					]
				}
			],
			"relations": [
				{
					"type": "flights",
					"depth": {
						"min": 1,
						"max": 1
					},
					"entityFrom": -1,
					"entityTo": 0,
					"constraints": []
				}
			],
			"limit": 5000
		}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	_, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.Equal(t, errors.New("non-existing entity referenced in return"), err)
}

func TestNoRelationsField(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
			"return": {
				"entities": [
					0
				]
			},
			"entities": [
				{
					"type": "airports",
					"constraints": [
						{
							"attribute": "city",
							"value": "San Francisco",
							"dataType": "string",
							"matchType": "exact"
						}
					]
				}
			],
			"limit": 5000
		}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	convertedResult, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)

	// Assert that the result and the expected result are the same
	correctConvertedResult := `MATCH (n0:airports)WHERE n0.city = "San Francisco" RETURN n0`
	cleanedResult := strings.ReplaceAll(*convertedResult, "\n", "")
	cleanedResult = strings.ReplaceAll(cleanedResult, "\t", "")
	assert.Equal(t, correctConvertedResult, cleanedResult)
}

func TestEntityFromLowerThanNegativeOneInRelation(t *testing.T) {
	// Setup for test
	// Create query conversion service
	service := NewService()

	query := []byte(`{
			"return": {
				"entities": [
					0
				],
				"relations": [
					0
				]
			},
			"entities": [
				{
					"type": "airports",
					"constraints": [
						{
							"attribute": "city",
							"value": "San Francisco",
							"dataType": "string",
							"matchType": "exact"
						}
					]
				}
			],
			"relations": [
				{
					"type": "flights",
					"depth": {
						"min": 1,
						"max": 1
					},
					"entityFrom": -4,
					"entityTo": 0,
					"constraints": []
				}
			],
			"limit": 5000
		}`)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	_, err := service.ConvertQuery(&JSONQuery)

	// Assert that there is no error
	assert.NoError(t, err)
}
